# Copyright (c) 2017 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from ovsdbapp import event
from ovsdbapp.tests import base


class TestEvent(event.RowEvent):
    def __init__(self):
        super(TestEvent, self).__init__(
            (self.ROW_CREATE,), "FakeTable", (("col", "=", "val"),))

    def run(self):
        pass

    def matches(self):
        pass


class TestRowEvent(base.TestCase):
    def test_compare_stop_event(self):
        r = TestEvent()
        self.assertFalse((r, "fake", "fake", "fake") == event.STOP_EVENT)
